package com.codigoparallevar.minicards;

import android.annotation.TargetApi;
import android.content.Context;
import android.graphics.Color;
import android.os.Build;
import android.support.annotation.NonNull;

import com.codigoparallevar.minicards.parts.buttons.RoundButton;
import com.codigoparallevar.minicards.parts.samples.ColorBox;
import com.codigoparallevar.minicards.parts.samples.Placeholder;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartConnection;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.Tuple2;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;

public class CardFile {

    static final int DEFAULT_BACKGROUND_COLOR = Color.parseColor("#044563");
    static final String PATH_SEPARATOR = "/";
    @NonNull
    private final String cardsDirectory;

    @NonNull
    private String name;
    @NonNull
    private List<Part> parts;
    private int backgroundColor;
    @NonNull
    private List<PartConnection> connections;
    @NonNull
    private String path;

    public CardFile(String cardsDirectory) {
        this.cardsDirectory = cardsDirectory;
        this.parts = new ArrayList<>();
        this.connections = new ArrayList<>();
        this.name = generateAnonymousName();
        regeneratePath();
        this.backgroundColor = DEFAULT_BACKGROUND_COLOR;
    }

    private void regeneratePath() {
        this.path = cardsDirectory + PATH_SEPARATOR + name + ".json";
    }

    @NonNull
    private String generateAnonymousName() {
        return "unnamed-" + UUID.randomUUID().toString().substring(0, 6);
    }

    @NonNull
    public CardFile setName(String name) {
        this.name = name;
        regeneratePath();

        return this;
    }

    public void setPath(@NonNull String path) {
        this.path = path;
    }

    @NonNull
    public CardFile setBackgroundColor(int backgroundColor) {
        this.backgroundColor = backgroundColor;

        return this;
    }

    @NonNull
    public String getName() {
        return name;
    }

    int getBackgroundColor() {
        return backgroundColor;
    }

    @NonNull
    public List<Part> getParts() {
        return parts;
    }

    @NonNull
    public List<PartConnection> getConnections() {
        return connections;
    }

    @NonNull
    public String getPath() {
        return path;
    }

    @NonNull
    public CardFile addParts(List<Part> parts) {
        for (Part part : parts) {
            this.addPart(part);
        }

        return this;
    }

    @NonNull
    private CardFile addPart(Part part) {
        this.parts.add(part);

        return this;
    }

    private void addConnections(List<PartConnection> connections) {
        for (PartConnection connection : connections) {
            addConnection(connection);
        }
    }

    private void addConnection(PartConnection connection) {
        connections.add(connection);
    }

    public static void createDefaultCards(File cardsDir) throws IOException {
        // Create a button and a color box, and connect them
        RoundButton button = new RoundButton(new StubPartGrid(), 200, 400, 80, 100);
        ColorBox box = new ColorBox(new StubPartGrid(), 500, 350, 600, 450);
        button.getPressedOutputConnector().connectTo(box.getToggleInputConnector());

        CardFile buttonAndColorBoxCard = (new CardFile(cardsDir.getAbsolutePath())
                .addPart(button)
                .addPart(box)
                .setName("Button and color box"));

        String buttonAndColorBoxFilePath = (cardsDir.getAbsolutePath()
                + PATH_SEPARATOR + "Button and color box.json");

        buttonAndColorBoxCard.save(buttonAndColorBoxFilePath);
    }

    public void save(Context context) throws IOException {
        String cardPath = getDefaultCardStorage(context) + PATH_SEPARATOR + name + ".json";

        save(cardPath);
    }

    public void save(String cardPath) throws IOException {
        FileOutputStream fileOut = new FileOutputStream(cardPath);

        byte[] serialized = serializeState();
        fileOut.write(serialized);

        fileOut.close();
    }

    @NonNull
    private byte[] serializeState() throws IOException {
        final JSONArray partArray = new JSONArray();
        for (Part part : parts) {
            JSONObject serializedPart = new JSONObject();
            try {
                serializedPart.put("_type", part.getClass().getName());
                serializedPart.put("_data", part.serialize());
            } catch (JSONException e) {
                throw new IOException(e.getMessage(), e.getCause());
            }

            partArray.put(serializedPart);
        }

        final JSONObject metadataObject = new JSONObject(new HashMap<String, Object>(){{
            put("version", 0);
            put("name", name);
            put("backgroundColor", serializeColor(backgroundColor));
        }});

        final JSONObject state = new JSONObject(new HashMap<String, Object>(){{
            put("metadata", metadataObject);
            put("parts", partArray);
        }});
        return state.toString().getBytes("UTF-8");
    }

    public static String serializeColor(int color) {
        return String.format("#%02x%02x%02x",
                (color & 0xFF0000) >> 16,
                (color & 0x00FF00) >> 8,
                (color & 0x0000FF));
    }

    @TargetApi(Build.VERSION_CODES.KITKAT)
    @NonNull
    public static CardFile load(String path, PartGrid grid) throws ErrorLoadingCardException, IOException {
        CardFile card = new CardFile(new File(path).getParent());

        File file = new File(path);

        try (FileReader fileIn = new FileReader(file)) {
            char[] data = new char[(int) file.length()];
            fileIn.read(data);
            JSONObject state = new JSONObject(new String(data));
            JSONObject metadata = state.getJSONObject("metadata");
            int version = metadata.getInt("version");

            if (version != 0) {
                throw new ErrorLoadingCardException(ErrorLoadingCardException.Reason.UNKNOWN_VERSION);
            }

            card.setPath(path);
            card.setName(metadata.getString("name"));
            card.setBackgroundColor(Color.parseColor(metadata.getString("backgroundColor")));

            // If version 0 -> known structure
            JSONArray jsonParts = state.getJSONArray("parts");

            for (int i = 0; i < jsonParts.length(); i++) {
                Tuple2<Part, List<PartConnection>> info = deserializeObject(grid, jsonParts.getJSONObject(i));

                card.addPart(info.item1);
                card.addConnections(info.item2);
            }
        } catch (JSONException e) {
            throw new ErrorLoadingCardException(ErrorLoadingCardException.Reason.UNKNOWN_FORMAT);
        }

        return card;
    }

    @NonNull
    private static Tuple2<Part, List<PartConnection>> deserializeObject(PartGrid grid,
                                                          JSONObject jsonObject) throws JSONException {
        String type = jsonObject.getString("_type");

        if(type.equals(RoundButton.class.getName())) {
            Tuple2<Part, List<PartConnection>> buttonInfo = RoundButton.deserialize(
                    grid,
                    jsonObject.getJSONObject("_data"));

            return buttonInfo;
        }
        else if (type.equals(Placeholder.class.getName())) {
            return new Tuple2<>(Placeholder.deserialize(grid, jsonObject.getJSONObject("_data")),
                    Collections.<PartConnection>emptyList());
        }
        else if (type.equals(ColorBox.class.getName())){
            return new Tuple2<>(ColorBox.deserialize(grid, jsonObject.getJSONObject("_data")),
                    Collections.<PartConnection>emptyList());
        }
        else {
            throw new JSONException("Expected known class, found " + type);
        }
    }

    public static String getDefaultCardStorage(Context context) {
        return context.getFilesDir().getAbsolutePath() + CardFile.PATH_SEPARATOR + "cards";
    }
}
